# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from enum import Enum

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ---------------------------------------------------------- class: FunctionName --------------------------------------------------------- #

class FunctionName(Enum):
    HT_DCPERIOD         = 'HT_DCPERIOD'
    HT_DCPHASE          = 'HT_DCPHASE'
    HT_PHASOR           = 'HT_PHASOR'
    HT_SINE             = 'HT_SINE'
    HT_TRENDMODE        = 'HT_TRENDMODE'
    ADD                 = 'ADD'
    DIV                 = 'DIV'
    MAX                 = 'MAX'
    MAXINDEX            = 'MAXINDEX'
    MIN                 = 'MIN'
    MININDEX            = 'MININDEX'
    MINMAX              = 'MINMAX'
    MINMAXINDEX         = 'MINMAXINDEX'
    MULT                = 'MULT'
    SUB                 = 'SUB'
    SUM                 = 'SUM'
    ACOS                = 'ACOS'
    ASIN                = 'ASIN'
    ATAN                = 'ATAN'
    CEIL                = 'CEIL'
    COS                 = 'COS'
    COSH                = 'COSH'
    EXP                 = 'EXP'
    FLOOR               = 'FLOOR'
    LN                  = 'LN'
    LOG10               = 'LOG10'
    SIN                 = 'SIN'
    SINH                = 'SINH'
    SQRT                = 'SQRT'
    TAN                 = 'TAN'
    TANH                = 'TANH'
    ADX                 = 'ADX'
    ADXR                = 'ADXR'
    APO                 = 'APO'
    AROON               = 'AROON'
    AROONOSC            = 'AROONOSC'
    BOP                 = 'BOP'
    CCI                 = 'CCI'
    CMO                 = 'CMO'
    DX                  = 'DX'
    MACD                = 'MACD'
    MACDEXT             = 'MACDEXT'
    MACDFIX             = 'MACDFIX'
    MFI                 = 'MFI'
    MINUS_DI            = 'MINUS_DI'
    MINUS_DM            = 'MINUS_DM'
    MOM                 = 'MOM'
    PLUS_DI             = 'PLUS_DI'
    PLUS_DM             = 'PLUS_DM'
    PPO                 = 'PPO'
    ROC                 = 'ROC'
    ROCP                = 'ROCP'
    ROCR                = 'ROCR'
    ROCR100             = 'ROCR100'
    RSI                 = 'RSI'
    STOCH               = 'STOCH'
    STOCHF              = 'STOCHF'
    STOCHRSI            = 'STOCHRSI'
    TRIX                = 'TRIX'
    ULTOSC              = 'ULTOSC'
    WILLR               = 'WILLR'
    BBANDS              = 'BBANDS'
    DEMA                = 'DEMA'
    EMA                 = 'EMA'
    HT_TRENDLINE        = 'HT_TRENDLINE'
    KAMA                = 'KAMA'
    MA                  = 'MA'
    MAMA                = 'MAMA'
    # MAVP                = 'MAVP'  # NEEDS 'periods' SO IS NOT CURRENTLY SUPPORTED
    MIDPOINT            = 'MIDPOINT'
    MIDPRICE            = 'MIDPRICE'
    SAR                 = 'SAR'
    SAREXT              = 'SAREXT'
    SMA                 = 'SMA'
    T3                  = 'T3'
    TEMA                = 'TEMA'
    TRIMA               = 'TRIMA'
    WMA                 = 'WMA'
    CDL2CROWS           = 'CDL2CROWS'
    CDL3BLACKCROWS      = 'CDL3BLACKCROWS'
    CDL3INSIDE          = 'CDL3INSIDE'
    CDL3LINESTRIKE      = 'CDL3LINESTRIKE'
    CDL3OUTSIDE         = 'CDL3OUTSIDE'
    CDL3STARSINSOUTH    = 'CDL3STARSINSOUTH'
    CDL3WHITESOLDIERS   = 'CDL3WHITESOLDIERS'
    CDLABANDONEDBABY    = 'CDLABANDONEDBABY'
    CDLADVANCEBLOCK     = 'CDLADVANCEBLOCK'
    CDLBELTHOLD         = 'CDLBELTHOLD'
    CDLBREAKAWAY        = 'CDLBREAKAWAY'
    CDLCLOSINGMARUBOZU  = 'CDLCLOSINGMARUBOZU'
    CDLCONCEALBABYSWALL = 'CDLCONCEALBABYSWALL'
    CDLCOUNTERATTACK    = 'CDLCOUNTERATTACK'
    CDLDARKCLOUDCOVER   = 'CDLDARKCLOUDCOVER'
    CDLDOJI             = 'CDLDOJI'
    CDLDOJISTAR         = 'CDLDOJISTAR'
    CDLDRAGONFLYDOJI    = 'CDLDRAGONFLYDOJI'
    CDLENGULFING        = 'CDLENGULFING'
    CDLEVENINGDOJISTAR  = 'CDLEVENINGDOJISTAR'
    CDLEVENINGSTAR      = 'CDLEVENINGSTAR'
    CDLGAPSIDESIDEWHITE = 'CDLGAPSIDESIDEWHITE'
    CDLGRAVESTONEDOJI   = 'CDLGRAVESTONEDOJI'
    CDLHAMMER           = 'CDLHAMMER'
    CDLHANGINGMAN       = 'CDLHANGINGMAN'
    CDLHARAMI           = 'CDLHARAMI'
    CDLHARAMICROSS      = 'CDLHARAMICROSS'
    CDLHIGHWAVE         = 'CDLHIGHWAVE'
    CDLHIKKAKE          = 'CDLHIKKAKE'
    CDLHIKKAKEMOD       = 'CDLHIKKAKEMOD'
    CDLHOMINGPIGEON     = 'CDLHOMINGPIGEON'
    CDLIDENTICAL3CROWS  = 'CDLIDENTICAL3CROWS'
    CDLINNECK           = 'CDLINNECK'
    CDLINVERTEDHAMMER   = 'CDLINVERTEDHAMMER'
    CDLKICKING          = 'CDLKICKING'
    CDLKICKINGBYLENGTH  = 'CDLKICKINGBYLENGTH'
    CDLLADDERBOTTOM     = 'CDLLADDERBOTTOM'
    CDLLONGLEGGEDDOJI   = 'CDLLONGLEGGEDDOJI'
    CDLLONGLINE         = 'CDLLONGLINE'
    CDLMARUBOZU         = 'CDLMARUBOZU'
    CDLMATCHINGLOW      = 'CDLMATCHINGLOW'
    CDLMATHOLD          = 'CDLMATHOLD'
    CDLMORNINGDOJISTAR  = 'CDLMORNINGDOJISTAR'
    CDLMORNINGSTAR      = 'CDLMORNINGSTAR'
    CDLONNECK           = 'CDLONNECK'
    CDLPIERCING         = 'CDLPIERCING'
    CDLRICKSHAWMAN      = 'CDLRICKSHAWMAN'
    CDLRISEFALL3METHODS = 'CDLRISEFALL3METHODS'
    CDLSEPARATINGLINES  = 'CDLSEPARATINGLINES'
    CDLSHOOTINGSTAR     = 'CDLSHOOTINGSTAR'
    CDLSHORTLINE        = 'CDLSHORTLINE'
    CDLSPINNINGTOP      = 'CDLSPINNINGTOP'
    CDLSTALLEDPATTERN   = 'CDLSTALLEDPATTERN'
    CDLSTICKSANDWICH    = 'CDLSTICKSANDWICH'
    CDLTAKURI           = 'CDLTAKURI'
    CDLTASUKIGAP        = 'CDLTASUKIGAP'
    CDLTHRUSTING        = 'CDLTHRUSTING'
    CDLTRISTAR          = 'CDLTRISTAR'
    CDLUNIQUE3RIVER     = 'CDLUNIQUE3RIVER'
    CDLUPSIDEGAP2CROWS  = 'CDLUPSIDEGAP2CROWS'
    CDLXSIDEGAP3METHODS = 'CDLXSIDEGAP3METHODS'
    AVGPRICE            = 'AVGPRICE'
    MEDPRICE            = 'MEDPRICE'
    TYPPRICE            = 'TYPPRICE'
    WCLPRICE            = 'WCLPRICE'
    BETA                = 'BETA'
    CORREL              = 'CORREL'
    LINEARREG           = 'LINEARREG'
    LINEARREG_ANGLE     = 'LINEARREG_ANGLE'
    LINEARREG_INTERCEPT = 'LINEARREG_INTERCEPT'
    LINEARREG_SLOPE     = 'LINEARREG_SLOPE'
    STDDEV              = 'STDDEV'
    TSF                 = 'TSF'
    VAR                 = 'VAR'
    ATR                 = 'ATR'
    NATR                = 'NATR'
    TRANGE              = 'TRANGE'
    AD                  = 'AD'
    ADOSC               = 'ADOSC'
    OBV                 = 'OBV'


# ---------------------------------------------------------------------------------------------------------------------------------------- #


# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from typing import Optional, List, Dict, Union, Any

# Pip
from jsoncodable import JSONCodable

from talib import abstract
import numpy

# Local
from .enums.function_name import FunctionName

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------------------ Public methods ------------------------------------------------------------ #

__ALL_METHODS_KEY = 'ALL'
ALL = __ALL_METHODS_KEY

# Supports 'ALL' as type
def get_stats(
    data: Dict[str, numpy.ndarray],
    types: Union[List[FunctionName], str] = __ALL_METHODS_KEY,
    timeperiod: int = 30,
    use_builtin_types: bool = False,
    use_objects: bool = False
) -> Optional[Union[Dict[str, Any], object]]:
    """Creates TA data for the selected types of analysises

    Args:
        data (Dict[str, numpy.ndarray]): The data to analyze. Should be provided in the following format:
        {
            'open': numpy.ndarray,
            'high': numpy.ndarray,
            'low': numpy.ndarray,
            'close': numpy.ndarray,
            'volume': numpy.ndarray,
        }

        types ([Union[List[FunctionName], str]]): The analysis types, you want to perform.
                                                  Defaults to ALL.

        timeperiod (int, optional): Time period to calculate analysis with.
                                    Defaults to 30.

        use_builtin_types (bool, optional): returns python list and int/float instead of numpy types.
                                            Defaults to False.

        use_objects (bool, optional): returns oject instead of Dict.
                                      If True, 'use_builtin_types' will be also True.
                                      Defaults to False.

        IMPORTANT NOTE:
            - THE VALUES OF 'use_builtin_types' AND 'use_objects' HIGHLY AFFECTS EFFICIENCY
              SEE BENCHMARK BELOW (Ran 250 times)

            --------------------------------------------------------------------------
            | rank |                    name                    |  duration  |  perc |
            --------------------------------------------------------------------------
            |    1 | use_builtin_types=False, use_objects=False | 0.00626893 |       |
            |    2 |  use_builtin_types=True, use_objects=False | 0.02440002 | 3.89x | -> ~4x slower
            |    3 |   use_builtin_types=True, use_objects=True | 0.03171391 | 5.05x | -> ~5x slower
            --------------------------------------------------------------------------

    Returns:
        Optional[Union[Dict[str, Any], object]]: Returns dictionary or objects, based on the provided values for 'use_builtin_types' and 'use_objects'
    """
    if isinstance(types, str):
        if types.upper() == __ALL_METHODS_KEY:
            types = [function_name for function_name in FunctionName]
        else:
            return None

    func_names = [function_name.value for function_name in types]
    res = {}

    for func_name in func_names:
        func = abstract.Function(func_name)

        try:
            _res = func(data, timeperiod=timeperiod)

            if use_builtin_types or use_objects:
                _res = __convert_to_builtin(_res)

            res[func_name] = _res
        except Exception as e:
            print(func_name, e)

    return res if not use_objects else JSONCodable.from_json(res)


# ----------------------------------------------------------- Private methods ------------------------------------------------------------ #

def __convert_to_builtin(numpy_list) -> List:
    new_list = []

    for element in numpy_list:
        if isinstance(element, numpy.ndarray):
            element = __convert_to_builtin(element)
        elif numpy.isnan(element):
            element = None
        else:
            element = element.item()

        new_list.append(element)

    return new_list


# ---------------------------------------------------------------------------------------------------------------------------------------- #